/*
 * Copyright (C) 1996-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
/*-
 * Copyright (c) 1997 Niklas Hallqvist.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
__RCSID("$LAAS: terrain.c,v 1.12 2003/06/20 20:05:27 matthieu Exp $");

/***
 *** Gestion de l'affichage de Modeles Numeriques de Terrain (MNT).
 ***/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/gl.h>
#include <tcl.h>
#define GDHE_PRIVATE
#include "gdhe_terrain.h"
#include "vecteurs.h"
#include "terrain.h"
#include "global.h"
#include "math.h"

/* define this macro to return false (zero) to disable black strips */
#define STRIP_SIZE(x) 0
/* #define STRIP_SIZE(x) ((fabs((fabs(x)-(int)(fabs(x)+0.5)))<=0.05)?1:0) */

static GDHE_MNT_INT *convertGdheMnt(GDHE_MNT *mnt);
static GDHE_MNT *convertGdheMntInt(GDHE_MNT_INT *mnt);
static void mntColor(Tcl_Interp *interp, int type, unsigned char color[]);
static void flushMntColor(void);

/* Cache des couleurs du MNT */
static unsigned char **colorCache = NULL;

#define MNT_EPSILON 0.000001

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif 

#ifndef swap16
#define swap16(x)						      \
    (uint16_t)(((uint16_t)(x) & 0xff) << 8 | ((uint16_t)(x) & 0xff00) >> 8)
#endif 

#ifndef swap32
#define swap32(x)                                                     \
    (uint32_t)(((uint32_t)(x) & 0xff) << 24 |                         \
    ((uint32_t)(x) & 0xff00) << 8 | ((uint32_t)(x) & 0xff0000) >> 8 | \
    ((uint32_t)(x) & 0xff000000) >> 24)
#endif

/*----------------------------------------------------------------------*/
/*
 * Swap bytes in a double - more portable 
 * than the extension of above macros
 */
static double
swapDouble(double x)
{
	union dc { 
		double r;
		char p[8];
	} *tmp1, tmp2;
	int i;

	tmp1 = (union dc *)&x;

	for (i = 0; i < 8; i++) {
		tmp2.p[i] = tmp1->p[7-i];
	}
	return(tmp2.r);
}

/*----------------------------------------------------------------------*/
/**
 ** Read the header of a data file independantly of its byte order 
 **/
static int
readTerrainHeader(FILE *in, GDHE_MNT *mnt, int *swapit)
{
	GDHE_MNT_HEADER h;
	int  type;

	if (fread(&h, sizeof(GDHE_MNT_HEADER), 1, in) != 1) {
		return -1;
	}
	type = h.storageType;
	if (type == GDHE_MNT_X_COLUMN || type == GDHE_MNT_X_ROW) {
		*swapit = FALSE;
	} else {
		type = swap32(type);
		if (type == GDHE_MNT_X_COLUMN || type == GDHE_MNT_X_ROW) {
			*swapit = TRUE;
		} else {
			/* Unknown byte order or corrupted file */
			fprintf(stderr, 
			    "Unknown byte order or corrupted file");
			return -1;
		}
	}
	if (*swapit) {
		mnt->hdr.x0 = swapDouble(h.x0);
		mnt->hdr.y0 = swapDouble(h.y0);
		mnt->hdr.z0 = swapDouble(h.z0);
		mnt->hdr.dx = swapDouble(h.dx);
		mnt->hdr.dy = swapDouble(h.dy);
		mnt->hdr.dz = swapDouble(h.dz);
		mnt->hdr.nrows = swap32(h.nrows);
		mnt->hdr.ncols = swap32(h.ncols);
		mnt->hdr.storageType = swap32(h.storageType);
	} else {
		mnt->hdr.x0 = h.x0;
		mnt->hdr.y0 = h.y0;
		mnt->hdr.z0 = h.z0;
		mnt->hdr.dx = h.dx;
		mnt->hdr.dy = h.dy;
		mnt->hdr.dz = h.dz;
		mnt->hdr.nrows = h.nrows;
		mnt->hdr.ncols = h.ncols;
		mnt->hdr.storageType = h.storageType;
	}
	return 0;
}

/*----------------------------------------------------------------------*/
/**
 ** Read one point from a data file, swapping data if needed
 **/
static int
readTerrainPoint(FILE *in, GDHE_MNT_POINT *p, int swapit)
{
	GDHE_MNT_POINT tmp;

	if (fread(&tmp, sizeof(GDHE_MNT_POINT), 1, in) != 1) {
		return -1;
	}
	if (swapit) {
		p->z = swap16(tmp.z);
		p->type = swap16(tmp.type);
	} else {
		p->z = tmp.z;
		p->type = tmp.type;
	}
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** Display a numerical terrain
 **/
int
terrainCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    int i, j, k;
    Tcl_HashEntry *entry;
    GDHE_MNT_INT *mnt;
    GDHE_MNT_POINT_INT *p1;
    GDHE_MNT_POINT_INT *p2;
    GDHE_MNT_POINT_INT p;
    unsigned char color[3];
    double	intpart;	/* tmp value */
    int		iMin, iMax;
    int		jMin, jMax;	/* Bounds the area that contains data points */
    int		stripOn;
    int		filled = 0;

    /* Parameters passed to the fct : */
    int		crop;		/* Indicates wheter the display is cropped or
				   not */ 
    int		lightDpy;	/* Indicated whether the ground is displayed
				   or not */

    if (argc != 4) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
		      "\"terrain id\" \"crop\" \"lightDpy\" ", -1));
	return TCL_ERROR;
    }
    /* Recherche du MNT a dessiner */
    entry = Tcl_FindHashEntry(mntTable, Tcl_GetStringFromObj(argv[1], NULL));
    if (entry == NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("cant find terrain", -1));
	return TCL_ERROR;
    }
    mnt = (GDHE_MNT_INT *)Tcl_GetHashValue(entry);
    /* Get the two remaining paramters */
    if (Tcl_GetIntFromObj(interp, argv[2], &crop) != TCL_OK)
      return TCL_ERROR;
    if (Tcl_GetIntFromObj(interp, argv[3], &lightDpy) != TCL_OK)
      return TCL_ERROR;
    
    flushMntColor();

    /* Let's go. First get the bounds */
    if (crop != 0) {
      iMin = mnt->nrows;
      iMax = 0;
      jMin = mnt->ncols;
      jMax = 0;
      p1 = mnt->points;
      for (i = 0; i < mnt->nrows; i++) {
	for (j = 0; j < mnt->ncols; j++) {
	  if (fabs(p1->point[2] - mnt->z0) > MNT_EPSILON)  {
	    /* TODO : one must do this because when saved from a VxWorks CPU,
	       the strict equality is not true... One should use 'type'
	       instead */ 
	    filled = 1;
	    iMin = ((iMin < i) ? iMin : i);
	    iMax = ((iMax > i) ? iMax : i);
	    jMin = ((jMin < j) ? jMin : j);
	    jMax = ((jMax > j) ? jMax : j);
	  }
	  p1++;
	}
      }
      if (filled == 0) {	/* Display nothing */
	iMin = iMax = 0;
	jMin = jMax = 0;
      }
    }
    else {
      iMin = 0;
      iMax = mnt->nrows;
      jMin = 0;
      jMax = mnt->ncols;
    }

    /* Border of terrain */
#if 0
    color[0] = 255; color[1] = color[2] = 0;
    glBegin(GL_LINE_LOOP);
    glColor3ubv(color);
    p = mnt->points[0]; p.point[2] = 0.0;
    glVertex3fv(p.point);
    p = mnt->points[mnt->ncols-1]; p.point[2] = 0.0;
    glVertex3fv(p.point);
    p = mnt->points[mnt->nrows*mnt->ncols-1]; p.point[2] = 0.0;
    glVertex3fv(p.point);
    p = mnt->points[(mnt->nrows-1)*mnt->ncols]; p.point[2] = 0.0;
    glVertex3fv(p.point);
    glEnd();
#endif

    /* Now, let's really go. */
    for (i = iMin; i < iMax - 1; i++) {
	k = i*mnt->ncols + jMin;
	stripOn = 0;
	if (lightDpy == 0) 
	  glBegin(GL_QUAD_STRIP);
	for (j = jMin; j < jMax - 1; j++) {

	  p1 = &(mnt->points[k]);
	  p2 = &(mnt->points[k+mnt->ncols]);

	  if ((lightDpy != 0)
	      && (stripOn == 0)
	      && (fabs(p1->point[2] - mnt->z0) > MNT_EPSILON)
 	      && (fabs(p2->point[2] - mnt->z0) > MNT_EPSILON)
	      && (fabs((p1 + 1)->point[2] - mnt->z0) > MNT_EPSILON)
	      && (fabs((p2 + 1)->point[2] - mnt->z0) > MNT_EPSILON)) {
	    glBegin(GL_QUAD_STRIP);
	    stripOn = 1;
	  }

	  if ((lightDpy == 0) || (stripOn != 0)) {
	    if (STRIP_SIZE(p1->point[0]) || STRIP_SIZE(p1->point[1])) {
	       color[0] = color[1] = color[2] = 0.0;
	    } else
	       mntColor(interp, p1->type, color);

	    glColor3ubv(color);
	    glNormal3fv(p1->normal);
	    glVertex3fv(p1->point);
	    
	    if (STRIP_SIZE(p2->point[0]) || STRIP_SIZE(p2->point[1])) {
	       color[0] = color[1] = color[2] = 0.0;
	    } else
	       mntColor(interp, p2->type, color);

	    glColor3ubv(color);
	    glNormal3fv(p2->normal);
	    glVertex3fv(p2->point);
	  }

	  if (lightDpy != 0) {
	     if ((fabs((p1 + 1)->point[2] - mnt->z0) < MNT_EPSILON)
		 || (fabs((p2 + 1)->point[2] - mnt->z0) < MNT_EPSILON)) {
		if (stripOn != 0)
		   glEnd();
		stripOn = 0;
	     }
	     if (j>=jMax-2 && stripOn) {
		glEnd();
		stripOn = 0;
	     }
	  }

	  k++;
	} /* for */
	if (lightDpy == 0)
	  glEnd();
    } /* for */
    
    Tcl_SetObjResult(interp, nullObj);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** read a terrain data file
 **/
int 
readTerrainCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    FILE *f;
    GDHE_MNT mnt;
    int i, n, swapit;

    if (argc != 3) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be"
		      "\"readTerrain name file\"", -1));
	return TCL_ERROR;
    }
    if ((f = fopen(Tcl_GetStringFromObj(argv[2], NULL), "r")) == NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("cannot open file", -1));
	return TCL_ERROR;
    }
    if (readTerrainHeader(f, &mnt, &swapit) != 0) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("problems reading file", -1));
	return TCL_ERROR;
    }
    n = mnt.hdr.ncols*mnt.hdr.nrows;
    mnt.points = (GDHE_MNT_POINT *)Tcl_Alloc(n*sizeof(GDHE_MNT_POINT));
    if (mnt.points == NULL) {
	return TCL_ERROR;
    }
    for (i = 0; i < n; i++) {
	    if (readTerrainPoint(f, mnt.points + i, swapit) != 0) {
		    Tcl_SetObjResult(interp, 
			Tcl_NewStringObj("problems reading file", -1));
		    fclose(f);
		    return TCL_ERROR;
	    }
    }
    fclose(f);

    n =  defineTerrain(Tcl_GetStringFromObj(argv[1], NULL), &mnt);
    Tcl_Free((char *)mnt.points);
    return n;
}

/*----------------------------------------------------------------------*/

/**
 ** read a DEM using the GEROMS format 
 **/
int 
readTerrainGeromsCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{
    FILE *f;
    GDHE_MNT mnt;
    int i, n;
    double z;

    if (argc != 3) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be"
		      "\"readTerrain name file\"", -1));
	return TCL_ERROR;
    }
    if ((f = fopen(Tcl_GetStringFromObj(argv[2], NULL), "r")) == NULL) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("cannot open file", -1));
	return TCL_ERROR;
    }
    fscanf(f, "%d %d %lf %lf", &(mnt.hdr.nrows), &(mnt.hdr.ncols),
	   &(mnt.hdr.dy), &(mnt.hdr.dx));
    n = mnt.hdr.nrows*mnt.hdr.ncols;
    mnt.hdr.dz = 0.01;
    mnt.hdr.x0 = mnt.hdr.y0 = mnt.hdr.z0 = 0;
    mnt.hdr.storageType = GDHE_MNT_X_ROW;
    mnt.points = (GDHE_MNT_POINT *)Tcl_Alloc(sizeof(GDHE_MNT_POINT)*n);
    for (i = 0; i < n; i++) {
	fscanf(f, "%lf", &z);
	if (z > -90.0) {
	    mnt.points[i].z = 100*z;
	    mnt.points[i].type = 1;
	} else {
	    if (i > 0) {
		mnt.points[i].z = mnt.points[i-1].z;
	    } else {
		mnt.points[i].z = 0;
	    }
	    mnt.points[i].type = 0;
	}
    } /* for */
    fclose (f);

    n =  defineTerrain(Tcl_GetStringFromObj(argv[1], NULL), &mnt);
    Tcl_Free((char *)mnt.points);
    return n;
}

/*----------------------------------------------------------------------*/

/**
 ** Update an existing terrain by reading a file. The format of the file is
 ** the same as the one read by the fct 'readTerrainCallBack'.
 **/
int 
updateTerrainCallBack(ClientData clientData, Tcl_Interp *interp,
		      int argc, Tcl_Obj *const argv[])
{
  Tcl_HashEntry*	entry;	/*  */
  GDHE_MNT_INT*		mnt0Int; /* The existing internal dem */
  GDHE_MNT*		mnt0;	/* The same, converted as a GDHE_MNT */
  GDHE_MNT_POINT*	p0;	/* One of its points */
  FILE*			f;	/* The file */
  GDHE_MNT		mnt1;	/* The dem read on the file */
  GDHE_MNT_POINT*	p1;	/* One of its points */
  int			n;	/* Number of points in the read dem*/
  int			i0Mnt1;
  int			j0Mnt1;	/* The (int) coordinates of the first point of
				   'mnt1' in the mnt0 */ 
  int			i,j;	/* To loop */
  int			dj0, dj1;
  int			swapit;

  /* Params check */
    if (argc != 3) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be"
		      "\"updateTerrain name file\"", -1));
	return TCL_ERROR;
    }

    /* Get the existing dem */
    entry = Tcl_FindHashEntry(mntTable, Tcl_GetStringFromObj(argv[1], NULL));
    if (entry == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("cant find terrain", -1));
      return TCL_ERROR;
    }
    mnt0Int = (GDHE_MNT_INT *)Tcl_GetHashValue(entry);
    
    /* Convert it to an 'external dem' (structure GDHE_MNT) */
    mnt0 = convertGdheMntInt (mnt0Int);
    if (mnt0 == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("error while converting a dem", -1));
      return TCL_ERROR;
    }

    /* Read the file, fills 'mnt1' */
    if ((f = fopen(Tcl_GetStringFromObj(argv[2], NULL), "r")) == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("cannot open file", -1));
      return TCL_ERROR;
    }
    if (readTerrainHeader(f, &mnt1, &swapit) != 0) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("problems reading file", -1));
      return TCL_ERROR;
    }
    n = mnt1.hdr.ncols*mnt1.hdr.nrows;
    mnt1.points = (GDHE_MNT_POINT *)Tcl_Alloc(n*sizeof(GDHE_MNT_POINT));
    if (mnt1.points == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("allocation problem", -1));
      return TCL_ERROR;
    }
    for (i = 0; i < n; i++) {
	    if (readTerrainPoint(f, mnt1.points + i, swapit) != 0) {
		    Tcl_SetObjResult(interp, 
			Tcl_NewStringObj("problems reading file", -1));
		    fclose(f);
		    return TCL_ERROR;
	    }
    } /* for */
    fclose(f);

    /* Check if the dem 'mnt1' can "fit" in the existing dem 'mnt0'. */
    if ((mnt1.hdr.dx != mnt0->hdr.dx) 
	|| (mnt1.hdr.dy != mnt0->hdr.dy) 
	|| (mnt1.hdr.dz != mnt0->hdr.dz)
	|| (mnt1.hdr.storageType != mnt0->hdr.storageType)) {
      Tcl_SetObjResult(interp, 
		       Tcl_NewStringObj("file not compatible with the dem", -1));
      return TCL_ERROR;
    }
    
    /* Find the index of the top-left point of 'mnt1' in 'mnt0', check again
       the compatibility of the 2 dems. */
    /* TODO : vrifier qu'on est bien au pixel prs */
    i0Mnt1 = (int) (((mnt1.hdr.x0 - mnt0->hdr.x0) / mnt1.hdr.dx) + 0.5);
    j0Mnt1 = (int) (((mnt1.hdr.y0 - mnt0->hdr.y0) / mnt1.hdr.dy) + 0.5);
    
    if ((i0Mnt1 + mnt1.hdr.nrows > mnt0->hdr.nrows)
	|| (j0Mnt1 + mnt1.hdr.ncols > mnt0->hdr.ncols)) {
      Tcl_SetObjResult(interp, 
		       Tcl_NewStringObj("dem file too big for the current dem !", -1));
      return TCL_ERROR;
    }

    /* Everything is OK. Loop on the points of mnt1 to update the mnt0 */
    dj1 = mnt1.hdr.ncols;
    dj0 = mnt0->hdr.ncols;
    for (i = 0; i < mnt1.hdr.nrows; i++) {
      for (j = 0; j < mnt1.hdr.ncols; j++) {
	p1 = &(mnt1.points[i*dj1 + j]);
	if (p1->z != 0) {
	  /* Update the corresponding mnt0 point */
	  p0 = &(mnt0->points[(i+i0Mnt1)*dj0 + (j+j0Mnt1)]);
	  p0->z = p1->z;
	  p0->type = p1->type;
	}
      }
    }
    
    /* Free the original internal dem (the following 'defineTerrain' fct call
       will generate a new one) */

    n =  defineTerrain(Tcl_GetStringFromObj(argv[1], NULL), mnt0);

    /* Free the read dem and the converted dem */
    Tcl_Free ((char*) mnt0Int->points);
    Tcl_Free ((char*) mnt0Int);
    Tcl_Free ((char*) mnt1.points);
    Tcl_Free ((char*) mnt0->points);
    Tcl_Free ((char*) mnt0);

    return n;
}



/*----------------------------------------------------------------------*/

/**
 ** Creation d'un MNT a partir de donnees XDR 
 **/
int
defineTerrain(char *name, GDHE_MNT *terrain)
{
    GDHE_MNT_INT *mntInt;
    Tcl_HashEntry *entry;
    int newEntry;

    mntInt = convertGdheMnt(terrain);
    entry = Tcl_CreateHashEntry(mntTable, name, &newEntry);
    if (entry == NULL) {
	return TCL_ERROR;
    }
    Tcl_SetHashValue(entry, (ClientData)mntInt);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** destruction d'un MNT 
 **/
int
deleteTerrain(char *name)
{
    GDHE_MNT_INT *mntInt;
    Tcl_HashEntry *entry;
    
    entry = Tcl_FindHashEntry(mntTable, name);
    if (entry == NULL) {
	return TCL_ERROR;
    }
    mntInt = Tcl_GetHashValue(entry);
    Tcl_Free((char *)mntInt->points);
    Tcl_Free((char *)mntInt);
    Tcl_DeleteHashEntry(entry);

    return TCL_OK;
}
/*----------------------------------------------------------------------*/

/**
 ** destruction d'un MNT
 **/
int
deleteTerrainCallBack(ClientData clientData, Tcl_Interp *interp,
	    int argc, Tcl_Obj *const argv[])
{

    if (argc != 2) {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong # args: should be "
		      "\"deleteTerrain name\"", -1));
	return TCL_ERROR;
    }
    if (deleteTerrain(Tcl_GetStringFromObj(argv[1], NULL)) == TCL_OK) {
	Tcl_SetObjResult(interp, nullObj);
	return TCL_OK;
    } else {
	Tcl_SetObjResult(interp, Tcl_NewStringObj("cannot find terrain", -1));
	return TCL_ERROR;
    }
}

/*----------------------------------------------------------------------*/

/**
 ** Conversion d'un MNT au format externe vers le format interne utilise
 ** pour l'affichage
 **/
static GDHE_MNT_INT *
convertGdheMnt(GDHE_MNT *mnt)
{
    GDHE_MNT_INT *mnt1;
    GDHE_MNT_POINT_INT *p;
    GDHE_MNT_POINT *q;
    int i, j, n, ind, di, dj;
    float v[4][3],norm[3],res[3],z[4];

    if (mnt == NULL) {
	return NULL;
    }
    mnt1 = (GDHE_MNT_INT *)Tcl_Alloc(sizeof(GDHE_MNT_INT));
    if (mnt1 == NULL) {
	return NULL;
    }
    n = mnt->hdr.nrows * mnt->hdr.ncols;
    p = (GDHE_MNT_POINT_INT *)Tcl_Alloc(sizeof(GDHE_MNT_POINT_INT)*n);
    if (p == NULL) {
	Tcl_Free((char *)mnt1);
	return NULL;
    }
    di = mnt1->nrows = mnt->hdr.nrows;
    dj = mnt1->ncols = mnt->hdr.ncols;
    mnt1->x0 = mnt->hdr.x0;
    mnt1->y0 = mnt->hdr.y0;
    mnt1->z0 = mnt->hdr.z0;
    mnt1->dx = mnt->hdr.dx;
    mnt1->dy = mnt->hdr.dy;
    mnt1->dz = mnt->hdr.dz;

    mnt1->points = p;

    /* calcul des normales des facettes */
    /* stockage temporaire dans les points du MNT resultat */
    ind = 0;
    for (i = 0; i < mnt->hdr.nrows - 1; i++) {
	for (j = 0; j < mnt->hdr.ncols - 1; j++) {
	    z[0] = mnt->points[ind].z;
	    z[1] = mnt->points[ind+1].z;
	    z[2] = mnt->points[ind+dj+1].z;
	    z[3] = mnt->points[ind+dj].z;
	    
	    vecfill(0, mnt->hdr.dy, (z[1]-z[0])*mnt->hdr.dz, v[0]);
	    vecfill(mnt->hdr.dx, 0, (z[2]-z[1])*mnt->hdr.dz, v[1]);
	    vecfill(0, -mnt->hdr.dy, (z[3]-z[2])*mnt->hdr.dz, v[2]);
	    vecfill(-mnt->hdr.dx, 0, (z[0]-z[3])*mnt->hdr.dz, v[3]);
	    
	    vecfill(.0,.0,.0, norm);
	    crossprod(v[2], v[1], res);
	    unitize(res, res);
	    vecadd(res, norm, norm);
	    
	    crossprod(v[3], v[2], res);
	    unitize(res, res);
	    vecadd(res, norm, norm);
	    
	    crossprod(v[0], v[3], res);
	    unitize(res, res);
	    vecadd(res, norm, norm);

	    unitize(norm, mnt1->points[ind].point);
	    ind++;

	} /* for */
    } /* for */

    /* 2ieme passe - calcul des normales des sommets par moyennage 
    * des normales des facettes */

    for (i = 1; i < mnt->hdr.nrows - 1; i++) {
	ind = i*dj+1;
	for (j = 1; j < mnt->hdr.ncols - 1; j++) {
	    vecfill(0, 0, 0, norm);
	    vecadd(mnt1->points[ind].point, norm, norm);
	    vecadd(mnt1->points[ind-dj].point, norm, norm);
	    vecadd(mnt1->points[ind-1].point, norm, norm);
	    vecadd(mnt1->points[ind-dj-1].point, norm, norm);
	    unitize(norm, mnt1->points[ind].normal);
	    ind++;
	} /* for */
    } /* for */
    /* 1ere ligne */
    for (j = 1; j < mnt->hdr.ncols - 1; j++) {
	vecfill(0, 0, 0, norm);
	vecadd(mnt1->points[j].point, norm, norm);
	vecadd(mnt1->points[j-1].point, norm, norm);
	unitize(norm, mnt1->points[j].normal);
    } /* for */
    /* derniere ligne */
    for (j = 1; j < mnt->hdr.ncols - 1; j++) {
	ind = (di - 1)*dj + j;
	vecfill(0, 0, 0, norm);
	vecadd(mnt1->points[ind-dj].point, norm, norm);
	vecadd(mnt1->points[ind-dj-1].point, norm, norm);
	unitize(norm, mnt1->points[ind].normal);
    } /* for */
    /* 1ere colonne */
    for (i = 1; i < mnt->hdr.nrows - 1; i++) {
	ind = i*dj;
	vecfill(0, 0, 0, norm);
	vecadd(mnt1->points[ind].point, norm, norm);
	vecadd(mnt1->points[ind - dj].point, norm, norm);
	unitize(norm, mnt1->points[ind].normal);
    } /* for */
    /* derniere colonne */
    for (i = 1; i < mnt->hdr.nrows - 1; i++) {
	ind = i*dj + dj - 1;
	vecfill(0, 0, 0, norm);
	vecadd(mnt1->points[ind - 1].point, norm, norm);
	vecadd(mnt1->points[ind - dj - 1].point, norm, norm);
	unitize(norm, mnt1->points[ind].normal);
    }/* for */
    /* les 4 coins */
    veccopy(mnt1->points[0].point, mnt1->points[0].normal);
    veccopy(mnt1->points[dj-2].point, mnt1->points[dj-1].normal);
    veccopy(mnt1->points[(di-2)*dj].point, 
	    mnt1->points[(di-1)*dj].normal);
    veccopy(mnt1->points[(di-1)*dj - 2].point, 
	    mnt1->points[di*dj - 1].normal);

    /* Pour finir, calcul des points 3D */
    p = &(mnt1->points[0]);
    for (i = 0; i < mnt->hdr.nrows; i++) {
	for (j = 0;  j < mnt->hdr.ncols; j++) {
	    if (mnt->hdr.storageType == GDHE_MNT_X_ROW) {
		q = &(mnt->points[i*dj + j]);
		p->point[0] = mnt->hdr.x0 + i*mnt->hdr.dx;
		p->point[1] = mnt->hdr.y0 + j*mnt->hdr.dy;
		p->point[2] = mnt->hdr.z0 + q->z*mnt->hdr.dz;
	    } else {
		q = &(mnt->points[j*di + i]);
		p->point[0] = mnt->hdr.x0 + j*mnt->hdr.dx;
		p->point[1] = mnt->hdr.y0 + i*mnt->hdr.dy;
		p->point[2] = mnt->hdr.z0 + q->z*mnt->hdr.dz;
	    }
	    /* maj des dimensions de l'environnement */
	    if (p->point[0] < xmin) xmin = p->point[0];
	    if (p->point[1] < ymin) ymin = p->point[1];
	    if (p->point[2] < zmin) zmin = p->point[2];
	    if (p->point[0] > xmax) xmax = p->point[0];
	    if (p->point[1] > ymax) ymax = p->point[1];
	    if (p->point[2] > zmax) zmax = p->point[2];
	    
	    p->type = q->type;
	    p++;
	} /* for */
    } /* for */

    return mnt1;
} 

/**********************************************************************
 *
 * Conversion of an internal gdhe dem (GDHE_MNT_INT) to an external dem
 * (GDHE_MNT). This is required by the function updateTerrain, that updates an
 * existing GDHE_MNT_INT. 
 *
 **********************************************************************/

static GDHE_MNT*
convertGdheMntInt(GDHE_MNT_INT *mnt)
{
    GDHE_MNT*		mnt1;	/* What is returned */
    GDHE_MNT_POINT*	q;
    GDHE_MNT_POINT_INT*	p;
    int i, j, n, di, dj;

    if (mnt == NULL) {
	return NULL;
    }

    /* Alloc the mnt1 */
    mnt1 = (GDHE_MNT*) Tcl_Alloc (sizeof (GDHE_MNT));
    if (mnt1 == NULL) {
      return NULL;
    }

    /* Copy the header infos */
    mnt1->hdr.dx = mnt->dx;
    mnt1->hdr.dy = mnt->dy;
    mnt1->hdr.dz = mnt->dz;
    mnt1->hdr.x0 = mnt->x0;
    mnt1->hdr.y0 = mnt->y0;
    mnt1->hdr.z0 = mnt->z0;
    di = mnt1->hdr.nrows = mnt->nrows;
    dj = mnt1->hdr.ncols = mnt->ncols;
    mnt1->hdr.storageType = GDHE_MNT_X_ROW; /* TODO : this is the default */

    /* Allocate the points array */
    n = mnt->nrows * mnt->ncols;
    mnt1->points = (GDHE_MNT_POINT*) Tcl_Alloc(n * sizeof (GDHE_MNT_POINT));
    if (mnt1->points == NULL) {
	return NULL;
    }

    /* Now loop on the mnt points to instanciate the mnt1 points */
    /* TODO : optimise the points address */
    for (i = 0; i < mnt->nrows; i++) {
      for (j = 0;  j < mnt->ncols; j++) {
	p = &(mnt->points[i*dj + j]);
	q = &(mnt1->points[i*dj + j]);
	q->z = (short) (((p->point[2] - mnt->z0) / mnt->dz) + 0.5);
	q->type = p->type;
      }
    }

    return mnt1;
} 

/*----------------------------------------------------------------------*/
/*
 * Definition de la couleur d'un sommet a` partir de son type
 */

static void
mntColor(Tcl_Interp *interp, int type, unsigned char color[3])
{
    char buf[20];
    char *p;
    int argc;
    char **argv;
    unsigned char *c;

    color[0] = color[1] = color[2] = 100;
    if (type < 0 || type > 65535) {
	return;
    }
    if (colorCache[type] != NULL) {
	color[0] = colorCache[type][0];
	color[1] = colorCache[type][1];
	color[2] = colorCache[type][2];
	return;
    }
	
    sprintf(buf, "%d", type);
    p = Tcl_GetVar2(interp, "mntColor", buf, TCL_GLOBAL_ONLY);
    if (p == NULL) {
	return;
    } 
    if (Tcl_SplitList(interp, p, &argc, &argv) != TCL_ERROR) {
	if (argc == 3) {
	    c = (unsigned char *)Tcl_Alloc(3*sizeof(unsigned char));
	    c[0] = color[0] = atoi(argv[0]);
	    c[1] = color[1] = atoi(argv[1]);
	    c[2] = color[2] = atoi(argv[2]);
	    colorCache[type] = c;
	}
    }
    if (argv != NULL) {
	Tcl_Free((char *)argv);
    }
}

/*
 * Nettoye le cache des couleurs de MNT
 */
static void
flushMntColor(void)
{
    int i;

    if (colorCache != NULL) {
	for (i = 0; i < 65536; i++) {
	    if (colorCache[i] != NULL) {
		Tcl_Free((char *)colorCache[i]);
		colorCache[i] = NULL;
	    }
	}
    } else {
	colorCache = (unsigned char **)Tcl_Alloc(65536*sizeof(unsigned char *));
	memset(colorCache, 0, 65536*sizeof(unsigned char *));
    }
}

	    
/*----------------------------------------------------------------------*/

int
Terrain_Init(Tcl_Interp *interp)
{
    int code;

    code = Tcl_PkgProvide(interp, "Terrain", "1.0");
    if (code != TCL_OK) {
        return code;
    }

    Tcl_CreateObjCommand(interp, "terrain", terrainCallBack, NULL, NULL);
    Tcl_CreateObjCommand(interp, "readTerrain", readTerrainCallBack, 
		      NULL, NULL);
    Tcl_CreateObjCommand(interp, "readTerrainGeroms", readTerrainGeromsCallBack, 
		      NULL, NULL);
    Tcl_CreateObjCommand(interp, "updateTerrain", updateTerrainCallBack, 
		      NULL, NULL);
    Tcl_CreateObjCommand(interp, "deleteTerrain", deleteTerrainCallBack,
		      NULL, NULL);
    return TCL_OK;
}


